
/* ====================================================================
 * Copyright (c) 1995 Ben Laurie.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by Ben Laurie
 *    for use in the Apache-SSL HTTP server project."
 *
 * 4. The name "Apache-SSL Server" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Ben Laurie
 *    for use in the Apache-SSL HTTP server project."
 *
 * THIS SOFTWARE IS PROVIDED BY BEN LAURIE ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL BEN LAURIE OR
 * HIS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of patches to the Apache HTTP server interfacing it
 * to SSLeay.
 * For more information on Apache-SSL, contact Ben Laurie <ben@algroup.co.uk>.
 *
 * For more information on Apache see http://www.apache.org.
 *
 * For more information on SSLeay see http://www.psy.uq.oz.au/~ftp/Crypto/.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>

#include <err.h>
#include <x509.h>
#include <pem.h>

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"

/*
Apache SSL Support

Oct 95
*/

/*

This version only uses the default/environment settings for CA certificates,
as SSLeay doesn't (currently) have a method to reset them. This means that
a single instance of Apache can only have one set of CA certificates. This
limitation cannot easily be overcome, as there would be no way to synchronise
accept()s with the correct run of the server if multiple servers are run,
unless they use different ports, or Apache is modded to run individual accepts
for each virtual host.

*/

typedef enum
    {
    VERIFY_NONE=0, /* At present, NONE and OPTIONAL are the same! */
    VERIFY_OPTIONAL=1,
    VERIFY_REQUIRE=2,
    } VerifyType;

typedef struct
    {
    char *szSSLRoot;
    char *szCertificateFile;
    char *szKeyFile;
    char *szCACertificatePath;
    char *szCACertificateFile;

    char *szReqCiphers;

    char *szLogFile;
    FILE *fileLogFile;

    char *szErrorFile;
    FILE *fileErrorFile;

    int nVerifyDepth;
    VerifyType nVerifyClient;

    X509 *px509Certificate;
    RSA *prsaKey;

    SSL_CTX *pSSLCtx;

    BOOL bFakeBasicAuth;
    } SSLConfigRec;

typedef struct
{
  /* 1) If cipher is banned, refuse */
  /* 2) If RequiredCiphers is NULL, accept */
  /* 3) If the cipher isn't required, refuse */

  table *tbRequiredCiphers;
  table *tbBannedCiphers;

} SSLDirConfigRec;

extern module ssl_module;

static void SSLLogError(server_rec *s)
    {
    unsigned long l;
    char str[MAX_STRING_LEN];

    /* Print out error messages */

    while(l=ERR_get_error())
      {
	ERR_error_string(l,str);
	log_error(str,s);
      }
    }

int SSLFixups(request_rec *orig)
{
  SSLConfigRec *pConfig= get_module_config(orig->server->module_config,
				       &ssl_module);
  table *e = orig->subprocess_env;
  int keysize, secretkeysize;
  char buf[MAX_STRING_LEN];
  char *cipher = SSL_get_cipher(orig->connection->client.ssl);

  table_set(e, "HTTPS", "on");
  table_set(e, "HTTPS_CIPHER", cipher);

  /* Set the keysizes */
  /* Just check the cipher */
  if(!strcmp(cipher, SSL_TXT_NULL_WITH_MD5))
    {
      keysize = secretkeysize = 0;
    }
  else
    if(!strcmp(cipher, SSL_TXT_RC4_128_WITH_MD5))
      {
	keysize = secretkeysize = 128;
      }
  else
    if(!strcmp(cipher, SSL_TXT_RC4_128_EXPORT40_WITH_MD5))
      {
	keysize = 128; secretkeysize = 40;
      }
  else
    if(!strcmp(cipher, SSL_TXT_RC2_128_CBC_WITH_MD5))
      {
	keysize = secretkeysize = 128;
      }
  else
    if(!strcmp(cipher, SSL_TXT_RC2_128_CBC_EXPORT40_WITH_MD5))
      {
	keysize = 128;
	secretkeysize = 40;
      }
  else
    if(!strcmp(cipher, SSL_TXT_IDEA_128_CBC_WITH_MD5))
      {
	keysize = secretkeysize = 128;
      }
  else
    if(!strcmp(cipher, SSL_TXT_DES_64_CBC_WITH_MD5))
      {
	keysize = secretkeysize = 64;
      }
  else
    if(!strcmp(cipher, SSL_TXT_DES_64_CBC_WITH_SHA))
      {
	keysize = secretkeysize = 64;
      }
  else
    if(!strcmp(cipher, SSL_TXT_DES_192_EDE3_CBC_WITH_MD5))
      {
	keysize = secretkeysize = 192;
      }
  else
    if(!strcmp(cipher, SSL_TXT_DES_192_EDE3_CBC_WITH_SHA))
      {
	keysize = secretkeysize = 192;
      }
  else
    if(!strcmp(cipher, SSL_TXT_DES_64_CFB64_WITH_MD5_1))
      {
	keysize = secretkeysize = 64;
      }
  else
    if(!strcmp(cipher, SSL_TXT_NULL))
      {
	keysize = secretkeysize = 0;
      }

  sprintf(buf, "%d", keysize);
  table_set(e, "HTTPS_KEYSIZE", buf);

  sprintf(buf, "%d", secretkeysize);
  table_set(e, "HTTPS_SECRETKEYSIZE", buf);


  /* Set SSL_CLIENT_DN */
  if(orig->connection->client.szClientX509 != NULL)
    {
      table_set(e, "SSL_CLIENT_DN", orig->connection->client.szClientX509);
    }

  return OK;
}

int SSLLogger(request_rec *orig)
{
  SSLConfigRec *pConfig= get_module_config(orig->server->module_config,
				       &ssl_module);

  conn_rec *conn = orig->connection;
  struct tm *t; long timz; char sign;
  char str[MAX_STRING_LEN];
  char tstr[MAX_STRING_LEN];

  t = get_gmtoff(&timz);
  sign = (timz < 0 ? '-' : '+');
  if(timz < 0) 
    timz = -timz;

  strftime(tstr,MAX_STRING_LEN,"[%d/%b/%Y:%H:%M:%S",t);
  sprintf(&tstr[strlen(tstr)], " %c%02ld%02ld]", sign, timz/3600,
	    timz%3600);

  sprintf(str, "%s %s", tstr, SSL_get_cipher(conn->client.ssl));
  if(conn->client.nVerifyError)
    sprintf(str, "%s %d %s", str, conn->client.nVerifyError,
	    X509_cert_verify_error_string(conn->client.nVerifyError));
  else
    strcat(str, " - -");

  if(conn->client.szClientX509)
    sprintf(str, "%s \"%s\"",str,conn->client.szClientX509);
  else
    strcat(str, " -");

  strcat(str, "\n");
  fputs(str, pConfig->fileLogFile);
  fflush(pConfig->fileLogFile);
  return OK;
}

static void *CreateSSLDirConfig(pool *p, char *dummy)
{
  SSLDirConfigRec *rec = pcalloc(p, sizeof(SSLDirConfigRec));

  rec->tbRequiredCiphers = make_table(p, 4);
  rec->tbBannedCiphers = make_table(p, 4);

  return(rec);
}

static void *MergeSSLDirConfig(pool *p, void *basev, void *addv)
{
  SSLDirConfigRec *base = (SSLDirConfigRec *)basev;
  SSLDirConfigRec *add = (SSLDirConfigRec *)addv;
  SSLDirConfigRec *new =
    (SSLDirConfigRec *) palloc(p, sizeof(SSLDirConfigRec));

  new->tbRequiredCiphers = overlay_tables(p,
					  add->tbRequiredCiphers,
					  base->tbRequiredCiphers);
  new->tbBannedCiphers = overlay_tables(p,
					add->tbBannedCiphers,
					base->tbBannedCiphers);

  return(new);
}

static char *SSLRequireCipher(cmd_parms *cmd,
			      SSLDirConfigRec *rec,
			      char *cipher)
{
  table_set(rec->tbRequiredCiphers, cipher, "Required");
  return NULL;
}

static char *SSLBanCipher(cmd_parms *cmd,
			  SSLDirConfigRec *rec,
			  char *cipher)
{
  table_set(rec->tbBannedCiphers, cipher, "Banned");
  return NULL;
}

static int SSLCheckCipher(request_rec *r)
{
  char *cipher = SSL_get_cipher(r->connection->client.ssl);
  SSLDirConfigRec *rec = (SSLDirConfigRec *)
    get_module_config(r->per_dir_config, &ssl_module);

  /* Check to see if BannedCipher list has it */
  if(table_get(rec->tbBannedCiphers, cipher))
    {
      char *buf;

      buf = pstrcat(r->pool, "Cipher ", cipher, " is forbidden", NULL);
      log_reason(buf, r->filename, r);
      return FORBIDDEN;
    }
  /* Check to see if it is in Permitted list */
  if(table_get(rec->tbRequiredCiphers, cipher))
    {
      /* Allowed */
      return OK;
    }

  /* Check to see if permitted list is empty */
  if(rec->tbRequiredCiphers->nelts == 0)
    {
      return OK;
    }
  else
    {
      char *buf;
      
      buf = pstrcat(r->pool, "Cipher ", cipher, " is not on the permitted list", NULL);
      log_reason(buf, r->filename, r);
      return FORBIDDEN;
    }
}
  
static void *CreateSSLServerConfig(pool *p,server_rec *s)
    {
    SSLConfigRec *rec=pcalloc(p,sizeof(SSLConfigRec));
    
    rec->szCertificateFile=rec->szKeyFile=rec->szLogFile=NULL;
    rec->szSSLRoot=NULL;
    rec->szErrorFile=NULL;
    rec->szReqCiphers=NULL;
    rec->nVerifyDepth=0;
    rec->nVerifyClient=VERIFY_NONE;
    rec->px509Certificate=NULL;
    rec->prsaKey=NULL;
    rec->bFakeBasicAuth=FALSE;
    return rec;
    }

static void InitSSL()
    {
    char *CAfile=NULL,*CApath=NULL;

    SSL_load_error_strings();
    ERR_load_crypto_strings();
    SSL_debug("/tmp/ssldebug");
    }

/***
 *
 * tm2sec() from David Robinson drtr@esi.co.uk
 * (From mod_proxy.c in Apache 1.1)
 *
 ***/

static time_t
tm2sec(const struct tm *t)
{
  static const int dayoffs[12]={306,337,0,31,61,92,122,153,184,214,245,275};
  time_t ret;
  int year=t->tm_year, days;

  if (t->tm_mon < 2)
    year--;
  days = t->tm_mday - 1 + dayoffs[t->tm_mon];
  days += year * 365 + year/4 - year/100 + (year/100 + 3)/4 - 25508;
  days = ((days * 24 + t->tm_hour) * 60 + t->tm_min) * 60 + t->tm_sec;

  ret = (time_t) days;

  if (year < 69 || year > 138 || days < 0)
    return -1;
  else
    return ret;
}

static time_t
ASN1_UTCTIME_time_t(ASN1_UTCTIME *utc)
{
  struct tm tm, ltm;
  time_t ret;


  int i = strlen(utc);


  if(i < 10)
    {
      return(-1);
    }
  for(i = 0; i < 10; i++)
    if((utc[i] > '9') || (utc[i] < '0'))
      {
	return(-1);
      }

  tm.tm_year = (utc[0]-'0') * 10 + (utc[1]-'0');

  /* Deal with Year 2000 like eay did */
  if(tm.tm_year < 70)
    tm.tm_year += 100;

  tm.tm_mon = (utc[2]-'0') * 10 + (utc[3]-'0') - 1;
  tm.tm_mday = (utc[4]-'0') * 10 + (utc[5]-'0');
  tm.tm_hour = (utc[6]-'0') * 10 + (utc[7]-'0');
  tm.tm_min = (utc[8]-'0') * 10 + (utc[9]-'0');
  tm.tm_sec = (utc[10]-'0') * 10 + (utc[11]-'0');
  tm.tm_isdst = -1;

  /* Use David Robinson's tm2sec() */
  return(tm2sec(&tm));
}

static int
check_exec(char *program)
{
  struct stat st;

  if(stat(program, &st) == -1)
    {
      return(0);
    }
  if(S_ISREG(st.st_mode) && (st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
    {
      return(1);
    }
  else
    {
      return(0);
    }
}

static void
mail_warning(server_rec *s, int days)
{
  char *sm_path;
  FILE *sm;
  char buf[MAX_STRING_LEN];
  int pid, status;

  /* Check /usr/lib/sendmail, and then /usr/sbin/sendmail */
  sm_path = "/usr/lib/sendmail";

  if(!check_exec(sm_path))
    {
      sm_path = "/usr/sbin/sendmail";
      if(!check_exec(sm_path))
	{
	  return;
	}
    }

  if(!(pid = fork()))
    {
      fclose(stderr);
      fclose(stdout);
      sprintf(buf, "%s -oi -oem -t", sm_path);

      sm = popen(buf, "w");
      if(days > 0)
	{
	  fprintf(sm, "To: %s\n"
		  "Subject: WARNING: The certificate for %s expires in %d day(s)\n"
		  "\n",
		  s->server_admin,
		  s->server_hostname,
		  days);
	}
      else
	{
	  fprintf(sm, "To: %s\n"
		  "Subject: The certificate for %s has expired\n"
		  "\n",
		  s->server_admin,
		  s->server_hostname);
	}

      /* Mail the message */
      fprintf(sm,
"\n"
"----------------- WARNING -----------------\n"
"\n"
"Your certificate needs to be renewed. Use the\n"
"renewal program supplied with Stronghold to\n"
"generate a renewal CSR for submission to your\n"
"certificate authority.\n"
"\n"
"If the certificate expires browsers will not\n"
"be able to connect to your site using SSL.\n"
"\n"
"----------------- WARNING -----------------\n");
      pclose(sm);
      exit(0);
    }
  waitpid(pid, &status, 0);
}
  

static void
check_expiration(server_rec *s, X509 *cert)
{
  ASN1_UTCTIME *notAfter, *notBefore;
  time_t begin, end;
  time_t cur_time;
	  
  /* Spew the expiration date here */
  notAfter = cert->cert_info->validity->notAfter;
  notBefore = cert->cert_info->validity->notBefore;

  ASN1_UTCTIME_print(stderr, notBefore);
  fprintf(stderr, " UTC - ");
  ASN1_UTCTIME_print(stderr, notAfter);
  fprintf(stderr, " UTC\n");

  /* Now actually check to see if it is near expiration */
  begin = ASN1_UTCTIME_time_t(notBefore);
  if(begin == -1)
    {
      perror("Unable to get begin time");
    }
  end = ASN1_UTCTIME_time_t(notAfter);
  if(end == -1)
    {
      perror("Unable to get end time");
    }

  cur_time = time(NULL);
  if(cur_time == -1)
    {
      perror("Unable to get current time");
    }

  /* Check */
  if(begin > cur_time)
    {
      fprintf(stderr, "WARNING: Certificate not yet valid\n"); 
    }
  if(end < cur_time)
    {
      fprintf(stderr, "WARNING: Certificate EXPIRED\n"); 
      mail_warning(s, -1);
    }
  
    {
      int days = (end - cur_time) / (3600 * 24);
      if(days < 30)
	{
	  fprintf(stderr, "WARNING: Certificate expires in %d day(s)\n",
		  days);

	  /* Send a warning if days is 30, 14, or less than 7 */
	  if(days == 30 || days == 14 || days < 7)
	    {
	      mail_warning(s, days);
	    }
	}
    }
}
    

static void GetCertificateAndKey(pool *p,
				 server_rec *s,SSLConfigRec *pConfig)
    {
    char buf[1024],prompt[1024];
    int n;
    FILE *f;
    char *szPath;

    if(pConfig->px509Certificate)
	{
	fprintf(stderr,"Attempt to reinitialise SSL for server %s\n",s->server_hostname);
	return;
	}

    /* CA certificate configuration */
    if(pConfig->szCACertificatePath)
      {
	pConfig->szCACertificatePath = make_full_path(p,
						      pConfig->szSSLRoot,
						      pConfig->szCACertificatePath);
      }


    if(pConfig->szCACertificateFile)
      {
	pConfig->szCACertificateFile = make_full_path(p,
						      pConfig->szSSLRoot,
						      pConfig->szCACertificateFile);
      }

    pConfig->pSSLCtx = SSL_CTX_new();


    if(pConfig->szReqCiphers != NULL)
      {
	if(!SSL_CTX_set_cipher_list(pConfig->pSSLCtx, pConfig->szReqCiphers))
	  {
	    /* Set pref ciphers failed */
	    fprintf(pConfig->fileErrorFile, "unable to set ciphers\n");
	    ERR_print_errors(pConfig->fileErrorFile);
	    SSLLogError(s);
	  }
      }

    if(!SSL_load_verify_locations(pConfig->pSSLCtx,
				  pConfig->szCACertificateFile,
				  pConfig->szCACertificatePath))
      {
	fprintf(stderr, "error setting default verify locations\n");
	ERR_print_errors(stderr);
	exit(1);
      }

    pConfig->szCertificateFile = make_full_path(p,
						pConfig->szSSLRoot,
						pConfig->szCertificateFile);
    f=fopen(pConfig->szCertificateFile,"r");
    if(!f)
	{
	  fprintf(stderr,"Can't open certificate file %s\n",pConfig->szCertificateFile);
	  exit(1);
	}

    fprintf(stderr,"%s: ", s->server_hostname);
    pConfig->px509Certificate=X509_new();
    if(!PEM_read_X509(f,&pConfig->px509Certificate,NULL))
	{
	fprintf(stderr,"Error reading server certificate file %s\n",
		pConfig->szCertificateFile);
	ERR_print_errors(stderr);
	exit(1);
	}
    else
      {
	check_expiration(s, pConfig->px509Certificate);
      }
	
    fclose(f);

    if(pConfig->szKeyFile)
      {
	szPath = make_full_path(p,
				pConfig->szSSLRoot,
				pConfig->szKeyFile);
      }
    else
      {
	szPath = pConfig->szCertificateFile;
      }
/* Otherwise the path already contains the name of the certificate file */
    f=fopen(szPath,"r");
    if(!f)
	{
	fprintf(stderr,"Can't open key file ");
	perror(szPath);
	exit(1);
	}

    pConfig->prsaKey=RSA_new();

    if(!PEM_read_RSAPrivateKey(f,&pConfig->prsaKey,NULL))
	{
        fprintf(stderr,"Error reading private key file %s\n",szPath);
	ERR_print_errors(stderr);
	exit(1);
	}
    }

static void InitSSLServer(server_rec *s,pool *p)
    {
    InitSSL();
    for ( ; s ; s = s->next)
	{
	SSLConfigRec *pConfig=get_module_config(s->module_config,&ssl_module);

	if(!pConfig->szSSLRoot)
	  {
	    fprintf(stderr, "No SSLRoot set for server %s\n", s->server_hostname);
	    exit(1);
	  }
	if(!pConfig->szCertificateFile)
	    {
	    fprintf(stderr,"No SSL Certificate set for server %s\n",s->server_hostname);
	    exit(1);
	    }
	if(pConfig->nVerifyClient < 0 || pConfig->nVerifyClient > VERIFY_REQUIRE)
	    {
	    fprintf(stderr,"Bad value for SSLVerifyClient (%d)\n",pConfig->nVerifyClient);
	    exit(1);
	    }
	if(!pConfig->szLogFile)
	    {
	    fprintf(stderr,"Required SSLLogFile missing\n");
	    exit(1);
	    }

	pConfig->szLogFile = server_root_relative(p, pConfig->szLogFile);

	pConfig->fileLogFile=fopen(pConfig->szLogFile,"a");
	if(!pConfig->fileLogFile)
	    {
	      perror(pConfig->szLogFile);
	      exit(1);
	    }

	if(pConfig->szErrorFile != NULL)
	  {
	    pConfig->szErrorFile = server_root_relative(p, pConfig->szErrorFile);
	    pConfig->fileErrorFile=fopen(pConfig->szErrorFile, "a");
	    if(!pConfig->fileErrorFile)
	      {
		pConfig->fileErrorFile = pConfig->fileLogFile;
	      }
	  }
	else
	  {
	    pConfig->fileErrorFile = pConfig->fileLogFile;
	  }

	GetCertificateAndKey(p,s,pConfig);
      }
}

static char *set_server_string_slot (cmd_parms *cmd, char *struct_ptr, char *arg)
    {
    /* This one should be pretty generic... */
    char *pConfig=get_module_config(cmd->server->module_config,&ssl_module);
  
    int offset=(int)cmd->info; 
    *(char **)(pConfig+offset)=arg;
    return NULL;
    }

static char *set_server_int_slot (cmd_parms *cmd, char *struct_ptr, char *arg)
    {
  /* This one should be pretty generic... */
    char *pConfig=get_module_config(cmd->server->module_config,&ssl_module);
  
    int offset=(int)cmd->info; 
    *(int *)(pConfig+offset) = atoi(arg);
    return NULL;
    }

static char *set_server_bool_slot (cmd_parms *cmd, char *struct_ptr)
    {
  /* This one should be pretty generic... */
    char *pConfig=get_module_config(cmd->server->module_config,&ssl_module);
  
    int offset=(int)cmd->info; 
    *(BOOL *)(pConfig+offset)=TRUE;
    return NULL;
    }

command_rec ssl_cmds[]=
{
  /* Server level config */
  { "SSLRoot", set_server_string_slot,
      (void *)XtOffsetOf(SSLConfigRec,szSSLRoot),RSRC_CONF,TAKE1,
      "Root directory to find SSL-files" },
  { "SSLCertificateFile", set_server_string_slot,
      (void *)XtOffsetOf(SSLConfigRec,szCertificateFile),RSRC_CONF,TAKE1,
      "PEM certificate file" },
  { "SSLCertificateKeyFile",set_server_string_slot,
      (void *)XtOffsetOf(SSLConfigRec,szKeyFile),RSRC_CONF,TAKE1,
      "Certificate private key file (assumed to be SSLCertificateFile if absent)" },
  { "SSLCACertificatePath",set_server_string_slot,
    (void *)XtOffsetOf(SSLConfigRec,szCACertificatePath),RSRC_CONF,TAKE1,
    "CA Certificate path (taken from SSL_CERT_DIR if absent)" },
  { "SSLCACertificateFile",set_server_string_slot,
    (void *)XtOffsetOf(SSLConfigRec,szCACertificateFile),RSRC_CONF,TAKE1,
    "CA Certificate file (taken from SSL_CERT_FILE if absent)" },
  { "SSLVerifyDepth",set_server_int_slot,
      (void *)XtOffsetOf(SSLConfigRec,nVerifyDepth),RSRC_CONF,TAKE1,
      "Verify depth (default 0)" },
  { "SSLVerifyClient",set_server_int_slot,
      (void *)XtOffsetOf(SSLConfigRec,nVerifyClient),RSRC_CONF,TAKE1,
      "Verify client (0=no, 1=optional, 2=required" },
  { "SSLFakeBasicAuth",set_server_bool_slot,
      (void *)XtOffsetOf(SSLConfigRec,bFakeBasicAuth),RSRC_CONF,NO_ARGS,
      "Translate client X509 into a user name" },
  { "SSLLogFile",set_server_string_slot,
	(void *)XtOffsetOf(SSLConfigRec,szLogFile),RSRC_CONF,TAKE1,
	"Place to dump SSL connection information" },
  { "SSLErrorFile",set_server_string_slot,
      (void *)XtOffsetOf(SSLConfigRec, szErrorFile),RSRC_CONF,TAKE1,
      "Place to dump SSL messages that have no better home" },
  { "SSLPreferredCiphers",set_server_string_slot,
      (void *)XtOffsetOf(SSLConfigRec, szReqCiphers),RSRC_CONF,TAKE1,
      "Colon-delimited list of required ciphers (deprecated)" },
  { "SSLRequiredCiphers",set_server_string_slot,
      (void *)XtOffsetOf(SSLConfigRec, szReqCiphers),RSRC_CONF,TAKE1,
      "Colon-delimited list of required ciphers" },

  /* Directory level config */
  { "RequireCipher", SSLRequireCipher, NULL, OR_FILEINFO, ITERATE,
      "add a cipher to the list of required ciphers" },
  { "BanCipher", SSLBanCipher, NULL, OR_FILEINFO, ITERATE,
      "add a cipher to the list of banned ciphers" },
  { NULL },
};

static conn_rec *SSLVerifyConn;

static const char six2pr[64+1]=
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

void uuencode(char *szTo,const char *szFrom)
    {
    const unsigned char *s;

    for(s=(const unsigned char *)szFrom ; *s ; s+=3)
	{
	*szTo++=six2pr[s[0] >> 2];
	*szTo++=six2pr[(s[0] << 4 | s[1] >> 4)&0x3f];
	if(!s[0])
	    break;
	*szTo++=six2pr[(s[1] << 2 | s[2] >> 6)&0x3f];
	if(!s[1])
	    break;
	*szTo++=six2pr[s[2]&0x3f];
	if(!s[2])
	    break;
	}
    *szTo++='\0';
    }

void ApacheSSLSetupConn(server_rec *server,os_conn_rec *conn,int fd)
    {
    SSLConfigRec *pConfig=get_module_config(server->module_config,&ssl_module);

    conn->ssl=SSL_new(pConfig->pSSLCtx);
    SSL_set_fd(conn->ssl,fd);
  }

/*
Fake a Basic authentication from the X509 client certificate.

This must be run fairly early on to prevent a real authentication from
occuring, in particular it must be run before anything else that authenticates
a user.

This means that the Module statement for this module should be LAST in the
Configuration file.
*/

static int FakeBasicAuth(request_rec *r)
    {
    SSLConfigRec *pConfig=get_module_config(r->server->module_config,&ssl_module);
    char b1[MAX_STRING_LEN],b2[MAX_STRING_LEN];
  
    if(!pConfig->bFakeBasicAuth)
	return DECLINED;
    if(r->connection->user)
	return DECLINED;
    if(!r->connection->client.szClientX509)
	return DECLINED;
    /*
       Fake a password - which one would be immaterial, as, it seems, an empty
       password in the users file would match ALL incoming passwords, if only we
       were using the standard crypt library routine. Unfortunately, SSLeay
       "fixes" a "bug" in crypt and thus prevents blank passwords from working.
       (IMHO what they really fix is a bug in the users of the code - failing to
       program correctly for shadow passwords).
       We need, therefore, to provide a password. This password can be matched by
       adding the string "xxj31ZMTZzkVA" as the password in the user file.
*/
    sprintf(b1,"%s:password",r->connection->client.szClientX509);
    uuencode(b2,b1);
    sprintf(b1,"Basic %s",b2);
    table_set(r->headers_in,"Authorization",b1);
    return DECLINED;
    }

int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error)
    {
    char *s;
    SSLConfigRec *pConfig=get_module_config(SSLVerifyConn->server->module_config,&ssl_module);
  
    s=(char *)X509_NAME_oneline(X509_get_subject_name(xs));
    if(s == NULL)
	{
	ERR_print_errors(pConfig->fileErrorFile);
	return(0);
	}
    if(depth == 0)
	SSLVerifyConn->client.szClientX509=pstrdup(SSLVerifyConn->pool,s);

    fprintf(pConfig->fileErrorFile,"depth=%d %s\n",depth,s);
    free(s);
    if(error == VERIFY_ERR_UNABLE_TO_GET_ISSUER)
	{
	s=(char *)X509_NAME_oneline(X509_get_issuer_name(xs));
	if(s == NULL)
	    {
	    fprintf(pConfig->fileErrorFile,"verify error\n");
	    ERR_print_errors(pConfig->fileErrorFile);
	    SSLLogError(SSLVerifyConn->server);
	    return(0);
	    }
	fprintf(pConfig->fileErrorFile,"issuer= %s\n",s);
	free(s);
	}
    if(!ok)
	{
	fprintf(pConfig->fileErrorFile,"verify error:num=%d:%s\n",error,
		X509_cert_verify_error_string(error));
	SSLLogError(SSLVerifyConn->server);
	SSLVerifyConn->client.szClientX509=NULL;
	}
    if(depth >= pConfig->nVerifyDepth)
	{
	fprintf(pConfig->fileErrorFile,"Verify depth exceeded\n");
	log_error("Verify depth exceeded",SSLVerifyConn->server);
	ok=0;
	}
    fprintf(pConfig->fileErrorFile,"verify return:%d\n",ok);
    return(ok);
    }

static BOOL ApacheSSLSetCertStuff(conn_rec *conn)
    {
    SSLConfigRec *pConfig=get_module_config(conn->server->module_config,&ssl_module);
    SSL *con=conn->client.ssl;
    char *cert_file=pConfig->szCertificateFile;
    char *key_file=pConfig->szKeyFile;

    if(cert_file != NULL)
	{
	if(SSL_use_certificate(con,pConfig->px509Certificate) <= 0)
	    {
	    fprintf(pConfig->fileErrorFile,"unable to set certificate\n");
	    ERR_print_errors(pConfig->fileErrorFile);
	    SSLLogError(conn->server);
	    return FALSE;
	    }
	if(SSL_use_RSAPrivateKey(con,pConfig->prsaKey) <= 0)
	    {
	    fprintf(pConfig->fileErrorFile,"unable to set private key\n");
	    ERR_print_errors(pConfig->fileErrorFile);
	    SSLLogError(conn->server);
	    return FALSE;
	    }
	}
    return TRUE;
    }

SSL *ApacheSSLNew(server_rec *s)
{
  SSL *conn;
  SSLConfigRec *pConfig=get_module_config(s->module_config,&ssl_module);

  conn = SSL_new(pConfig->pSSLCtx);
  if(conn == NULL)
    return(conn);
  if(pConfig->szReqCiphers != NULL)
    {
      if(!SSL_set_cipher_list(conn, pConfig->szReqCiphers))
	{
	  /* Set pref ciphers failed */
	  fprintf(pConfig->fileErrorFile, "unable to set ciphers\n");
	  ERR_print_errors(pConfig->fileErrorFile);
	  SSLLogError(s);
	}
    }
  return(conn);
}

void ApacheSSLSetupVerify(conn_rec *conn)
    {
    SSLConfigRec *pConfig=get_module_config(conn->server->module_config,&ssl_module);
    int nVerify=0;

    conn->client.szClientX509=NULL;

    SSLVerifyConn=conn;

    switch(pConfig->nVerifyClient)
	{
    case VERIFY_REQUIRE:
	nVerify|=SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
	break;

    case VERIFY_OPTIONAL:
	nVerify|=SSL_VERIFY_PEER;
	break;
	}

    conn->client.nVerifyError=VERIFY_OK;

/* Why call this twice?? Ben */  
    SSL_set_verify(conn->client.ssl,nVerify,ApacheSSLVerifyCallback);
    if(!ApacheSSLSetCertStuff(conn))
	{
	fprintf(pConfig->fileErrorFile,"ApacheSSLSetCertStuff failed\n");
	log_error("ApacheSSLSetCertStuff failed",conn->server);
	exit(1);
	}
    SSL_set_verify(conn->client.ssl,nVerify,ApacheSSLVerifyCallback);
    }

int os_conn_vprintf(conn_rec *conn,const char *fmt,va_list args)
    {
    static char buf[MAX_STRING_LEN];
    int ret,len;

    vsprintf(buf,fmt,args);
    len=strlen(buf);
    ret=SSL_write(conn->client.ssl,buf,len);
    if(ret < len)
	{
	log_error("SSL_write failed",conn->server);
	SSLLogError(conn->server);
	exit(1);
	}
    return ret;
    }

int _os_conn_printf(conn_rec *conn,const char *fmt,...)
    {
    va_list args;
    int ret;
    
    va_start(args,fmt);
    ret=os_conn_vprintf(conn,fmt,args);
    va_end(args);

    return ret;
    }

int os_conn_puts(char *buf,conn_rec *conn)
    {
    return SSL_write(conn->client.ssl,buf,strlen(buf));
    }

int os_conn_putc(int c,conn_rec *conn)
    {
    char buf[1];

    buf[0]=c;
    return SSL_write(conn->client.ssl,buf,1);
    }

int os_conn_read(void *buf,int unit,int len,conn_rec *conn)
    {
    int ret;
    X509 *peer;
    char *str;

/* Dodgy - but SSLeay knows no better */
    SSLVerifyConn=conn;

    while(!SSL_is_init_finished(conn->client.ssl))
	{
	SSLConfigRec *pConfig=get_module_config(conn->server->module_config,&ssl_module);
      
	ret=SSL_accept(conn->client.ssl);
	if(ret <= 0)
	    {
	    log_error("SSL_accept failed",conn->server);
	    SSLLogError(conn->server);
/* treat as EOF for now... */
	    return 0;
	    }
	if(conn->client.nVerifyError != VERIFY_OK)
	    {
	    log_error("Verification failed",conn->server);
	    SSLLogError(conn->server);
/* treat as EOF for now... */
	    return 0;
	    }

	if(pConfig->nVerifyClient == VERIFY_REQUIRE && !conn->client.szClientX509)
	    {
	    log_error("No client certificate",conn->server);
	    SSLLogError(conn->server);
	    return 0;
	    }
	}

    ret=SSL_read(conn->client.ssl,buf,unit*len);

    SSLVerifyConn=NULL;

    if(ret < 0)
	{
	log_error("SSL_read failed",conn->server);
	SSLLogError(conn->server);
	}
    if(ret <= 0)
	return ret;
    return ret/unit;
    }

char *os_conn_gets(char *buf,int len,conn_rec *conn)
    {
    int n,ret;
    char locbuf[1];
    
    for(n=0 ; n < len-1 ; ++n)
	{
	ret=os_conn_read(locbuf,1,1,conn);
	if(ret == 0)
	    return NULL;
	if(ret < 0)
	    {
	    log_error("SSL_read failed",conn->server);
	    exit(1);
	    }
	if(locbuf[0] == '\n')
	    break;
	buf[n]=locbuf[0];
	}
    buf[n]='\0';
    return buf;
    }

module ssl_module = {
   STANDARD_MODULE_STUFF,
   InitSSLServer,		/* initializer */
   CreateSSLDirConfig,		/* dir config creater */
   MergeSSLDirConfig,		/* dir merger --- default is to override */
   CreateSSLServerConfig,	/* server config */
   NULL,			/* merge server config */
   ssl_cmds,			/* command table */
   NULL,			/* handlers */
   NULL,			/* filename translation */
   FakeBasicAuth,		/* check_user_id */
   NULL,			/* check auth */
   SSLCheckCipher,		/* check access */
   NULL,			/* type_checker */
   SSLFixups,			/* fixups */
   SSLLogger,			/* logger */
};
